#!/usr/bin/env python3

import json
import ucpeif.wireguardvpnif as w
import logging
import logging.handlers
import ipaddress
import time
from systemd.journal import JournalHandler
import vppapi_if.vppapi_if as vppif

from cryptography.fernet import Fernet
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
import base64

import edgility_db.edgility_db as db
from sysrepo_wg_concentrator import sysrepo_wg_concentrator as wg_api

class WireGuardConcentratorErr(Exception):
    def __init__(self, message, recover_f=lambda _: None):
        super(WireGuardConcentratorErr, self).__init__(message)

class WireGuardConcentratorClient():

    DEFAULT_KEEPALIVE_VALUE = 1
    NO_ERR_MSG = ''

    STATIC_POOL_NAME = 'static_clients_'
    DYNAMIC_POOL_NAME = 'dynamic_clients_'
    DYNAMIC_POOL_COLUMNS = [
            "public_key",
            "hostname",
            "peer_endpoint",
            "tunnel_ip",
            "allowed_ips",
            "password"
        ]
    STATIC_POOL_COLUMNS = [
            "hostname",
            "tunnel_ip"
    ]

    def __init__(self, hostname, password, public_key, conc_port, endpoint):
        self._hostname = hostname
        self._password = password
        self._public_key = public_key
        self._port = conc_port
        self._endpoint = endpoint
        self._client_allowed_ips = None
        self._client_tunnel_ip = None

        self._conc_ifname = None
        self._if_index = None
        self._conc_tunnel_ip = None
        self._conc_allowed_ips = None
        self._net_info = None

        self._db = db.EdgilityDatabase()

        self._last_peer_added = None
        self._last_route_added = None

        # Client reponse data
        self._client_data = {
            "tunnel-ip": '',
            "concentrator-public-key": '',
            "allowed-ips": '',
            "concentrator-tunnel-ip": ''
        }

        self._init_logger()

        # Real work happens here
        self._add_client()

    @property
    def conf_ifname(self):
        return self._conc_ifname

    @conf_ifname.setter
    def conf_ifname(self, value):
        self._conc_ifname = value

        self._dynamic_table_name = self.DYNAMIC_POOL_NAME + self._conc_ifname
        self._static_table_name = self.STATIC_POOL_NAME + self._conc_ifname

    @classmethod
    def WireGuardConcentratorClientStartup(cls, wg_if_idx, conc_ifname, conc_port):
        cself = cls.__new__(cls)
        super(WireGuardConcentratorClient, cself).__init__()

        cself._if_index = wg_if_idx
        cself._port = conc_port
        cself.conf_ifname = conc_ifname

        cself._db = db.EdgilityDatabase()

        cself._init_logger()

        try:
            cself._load_clients_on_startup()
        except WireGuardConcentratorErr as ex:
            # TODO: recover here
            cself._logger.error(f'Printing exception {ex}')

    def __str__(self):
        return json.dumps(self._client_data)

    def get_vpp_client(self):
        return vppif.VppApiInterface('wireguard-concentrator')

    def _add_client(self): 
        try:
            self._fetch_wireguard_concentrator_instance()
            self._verify_client_credentials()
            self._decrypt_client_public_key()
            self._fetch_client_data()
            self._add_client_in_vpp()
            self._register_client_in_database()                                     

        except WireGuardConcentratorErr as ex:
            # TODO: recover here
            self._logger.error(f'Printing exception {ex}')

    
    def _verify_client_credentials(self):
        sr_wgc = wg_api.SysrepoWireGuardConcentrator(app_name="wireguard concentrator")
        conc_password = sr_wgc.get_wireguard_concentrator_param(self._conc_ifname, 'concentrator-password')
        if not self._password == conc_password:
            raise WireGuardConcentratorErr(f"Error: Incorrect password!")

    def _fetch_wireguard_concentrator_instance(self):
        self.conf_ifname = self._get_concentrator_ifname()
        self._get_wg_interface_index()
        self._fetch_concentrator_tunnel_ip()

    def _get_concentrator_ifname(self):
        sr_wgc = wg_api.SysrepoWireGuardConcentrator(app_name="wireguard concentrator")
        interfaces = sr_wgc.get_wireguard_concentrator_interfaces()['wg-concentrator']
        for interface in interfaces:
            if interface.get('concentrator-port', None) == self._port:
                return interface.get('name', None)
        raise WireGuardConcentratorErr(f"Error: Interfaces {interfaces}")
    
    def _fetch_concentrator_tunnel_ip(self):
        vpp = self.get_vpp_client()
        self._conc_tunnel_ip = vpp.get_interface_address_by_ifidx(self._if_index)
        self._logger.error(f'conc tunnel ip {self._conc_tunnel_ip}')
        if not self._conc_tunnel_ip:
            raise WireGuardConcentratorErr(f"Error: No ip on interface with name {self._conc_ifname}")

    def _get_wg_interface_index(self):
        vpp = self.get_vpp_client()
        interface = vpp.wg_get_interface_by_local_port(self._port)
        self._if_index = interface['sw_if_index']
        del(vpp)
        if not self._if_index:
            raise WireGuardConcentratorErr(f"Error: Not existing wireguard inteface with local_port {self._port}")
    
    def _load_clients_on_startup(self):
        clients = self._db.fetch_all_rows_from_table(self._dynamic_table_name)
        for client in clients:
            self._public_key = self.extract_from_db_tuple_resp(client, 'public_key')
            self._endpoint = self.extract_from_db_tuple_resp(client, 'peer_endpoint')
            self._client_tunnel_ip = self.extract_from_db_tuple_resp(client, 'tunnel_ip')

            client_tunnel_ip_str = self.convert_ip_int_to_string(self._client_tunnel_ip)

            self._conc_allowed_ips = [f"{client_tunnel_ip_str}/32"]

            self._add_client_in_vpp()

    def _fetch_client_data(self):
        if self._fetch_client_data_on_existing_client():
            self._update_hostname_in_database() # This handles cases where client changed his hostname

            raise WireGuardConcentratorErr(f"Client data fetched from existing entry")

        self._client_tunnel_ip = self._get_tunnel_ip_from_static_pool() or self._get_tunnel_ip_from_dynamic_pool()
        if not self._client_tunnel_ip:
            raise WireGuardConcentratorErr(f"Error: Could not allocate tunnel_ip for client")

        conc_public_key = self._get_concentrator_public_key()
        if not conc_public_key:
            raise WireGuardConcentratorErr(f"Error: Could not get concentrator public key")

        client_tunnel_ip_str = self.convert_ip_int_to_string(self._client_tunnel_ip)
        self._conc_allowed_ips = [f"{client_tunnel_ip_str}/32"]
        # client_tunnel_netmask_str = str(self._net_info.prefixlen)

        self._logger.error(f"Client tunnel ip: {client_tunnel_ip_str}")

        self._client_data['tunnel-ip'] = f"{client_tunnel_ip_str}/32"
        self._client_data['concentrator-public-key'] = conc_public_key
        self._client_data['allowed-ips'] = self._client_allowed_ips[0]
        self._client_data['concentrator-tunnel-ip'] = str(self._conc_tunnel_ip).split('/')[0]

    def _fetch_client_data_on_existing_client(self):
        data = self._db.fetch_row_by_param(self._dynamic_table_name, 'public_key', self._public_key)
        self._logger.error(f'Existing client with data {data}')
        if data:
            client_tunnel_ip_int = self.extract_from_db_tuple_resp(data, "tunnel_ip")
            client_tunnel_ip_str = self.convert_ip_int_to_string(client_tunnel_ip_int)

            self._client_data['tunnel-ip'] = f'{client_tunnel_ip_str}/32'
            self._client_data['concentrator-public-key'] = self._get_concentrator_public_key()
            self._client_data['allowed-ips'] = self.extract_from_db_tuple_resp(data, 'allowed_ips')
            self._client_data['concentrator-tunnel-ip'] = str(self._conc_tunnel_ip).split('/')[0]

            return True
        return False

    def _get_tunnel_ip_from_static_pool(self):
        return self._db.fetch_from_table_by_param(self._static_table_name, 'tunnel_ip', 'hostname', self._hostname)
        
    def _get_tunnel_ip_from_dynamic_pool(self):
        tunnel_ips = self._db.fetch_column_entries(self._dynamic_table_name, 'tunnel_ip') 
        self._logger.error(f'Returning tunnel ip: {tunnel_ips}')
        return self._get_first_available_tunnel_ip(tunnel_ips)

    def _register_client_in_database(self):
        data_values = (str(self._public_key), str(self._hostname), str(self._endpoint), int(self._client_tunnel_ip), str(self._client_allowed_ips[0]), str(self._password))
        self._db.insert_values_in_table(self._dynamic_table_name, self.DYNAMIC_POOL_COLUMNS, data_values)

    def _update_hostname_in_database(self):
        self._db.update_column_value_on_matching_key(self._dynamic_table_name, 'hostname', self._hostname, 'public_key', self._public_key)

    def _get_first_available_tunnel_ip(self, ips):
        # Segment the length and check each one (for big ranges)
        self._net_info = ipaddress.IPv4Network(f"{self._conc_tunnel_ip}", strict=False)
        net_addr = int(self._net_info.network_address)
        available_hosts = self._net_info.num_addresses
        broadcast_addr = self._net_info.broadcast_address
        self._client_allowed_ips = [f'{self._net_info.network_address}/{self._net_info.prefixlen}']
        self._logger.error(f"Allowed ips: {self._client_allowed_ips}")

        if not ips:
            ips = []

        ips.append(int(self._conc_tunnel_ip))
        self._logger.error(int(self._conc_tunnel_ip))

        ips.sort()

        last_ip_idx = -1
        last_ip = ips[last_ip_idx]

        if available_hosts <= len(ips):
            return None

        while True:
            if last_ip - net_addr > len(ips[:ips.index(last_ip)]) + 1:
                if last_ip_idx == -len(ips):
                    return net_addr + 1
                last_ip_idx -= 1
                last_ip = ips[last_ip_idx]
                continue
            
            if last_ip + 1 == broadcast_addr:
                return None

            return last_ip + 1

    def _add_client_in_vpp(self):
        vpp = self.get_vpp_client()
        client_tunnel_ip_str = str(ipaddress.ip_address(self._client_tunnel_ip))
        ret, peer_idx = vpp.wg_add_peer(wgif=self._if_index, endpoint=self._endpoint, port=self._port, keepalive=self.DEFAULT_KEEPALIVE_VALUE,
                            allowed_ips=[f"{client_tunnel_ip_str}/32"], public_key_base64=self._public_key, base64=True)
        del(vpp)
        
        if ret:
            raise WireGuardConcentratorErr(f"Error executing wg_add_peer: {ret}")
        
        self._last_peer_added = peer_idx

        ret = self.add_del_peer_address_vpp_static_route(delete=False)
        if ret:
            raise WireGuardConcentratorErr(f"Error executing add_del_peer_address_vpp_sta   tic_route: {ret}")
        
    def add_del_peer_address_vpp_static_route(self, delete=False):
        netmask = 32
        
        for a_ip in self._conc_allowed_ips:
            ip_addr = ipaddress.IPv4Interface(str(a_ip))
            ip_netmask = ip_addr.network.prefixlen
            self._logger.error("WireGuardVPNInterface vpp_static_route ip_addr - {} ip_netmask {} - netmask {} - delete {}" .format(ip_addr, ip_netmask, netmask, delete))
            
            if netmask == ip_netmask:            
                vpp_if = self.get_vpp_client()
                if delete == True:
                    ret = vpp_if.del_ip_route('{}/{}' .format(ip_addr.ip, ip_netmask), next_hop_sw_if_index=self._if_index)
                    if ret:
                        self._logger.error ('Del ip route fail route {}/{} via {} ret={}'.format(ip_addr.ip, ip_netmask, self._if_index ,ret))
                    self._logger.error("WireGuardVPNInterface vpp_static_route static route for - {}/{} REMOVED" .format(ip_addr.ip, ip_netmask))
                else:
                    ret = vpp_if.add_ip_route('{}/{}' .format(ip_addr.ip, ip_netmask), next_hop_sw_if_index=self._if_index, check=False)
                    if ret:
                        self._logger.error ('Add ip route fail route {}/{} via {} ret={}'.format(ip_addr.ip, ip_netmask, self._if_index ,ret))
                    
                    self._last_route_added = f'{ip_addr.ip}/{ip_netmask}'

                    self._logger.error("WireGuardVPNInterface vpp_static_route static route for - {}/{} ADDED" .format(ip_addr.ip, ip_netmask))
                del(vpp_if)

                return ret
            
    def delete_single_vpp_route(self, route):
        vpp_if = self.get_vpp_client()
        ret = vpp_if.del_ip_route(route, next_hop_sw_if_index=self._if_index)
        if ret:
            self.logger.error(f'Error delete_single_vpp_route: {ret}')
        del(vpp_if)

    def _generate_key(self, password):
            kdf = PBKDF2HMAC(
                algorithm=hashes.SHA256(),
                length=32,
                salt=b'1V0eG071nP!cH',
                iterations=100000,
                backend=default_backend()
            )
            key = base64.urlsafe_b64encode(kdf.derive(password.encode()))
            return key
    
    def _decrypt_client_public_key(self):
        cipher = Fernet(self._generate_key(self._password))
        decrypted_public_key = cipher.decrypt(self._public_key.encode()).decode()
        self._public_key = decrypted_public_key
        if not self._public_key:
            raise WireGuardConcentratorErr(f"Error testing {decrypted_public_key}")

    def _get_concentrator_public_key(self):
        with open('/etc/wg/public.key', 'r') as f:
                public_key = f.read()

        return public_key

    def extract_from_db_tuple_resp(self, db_resp, value):
        index = self.DYNAMIC_POOL_COLUMNS.index(value)
        return db_resp[index]

    def convert_ip_int_to_string(self, ip_int):
        return str(ipaddress.ip_address(ip_int))

    def _init_logger(self, log_level=logging.DEBUG):
        self._logger = logging.getLogger(name=self.__class__.__name__)
        self._logger.setLevel(level=log_level)

        syslog_hnd = logging.handlers.SysLogHandler(address = '/dev/log')
        formatter = logging.Formatter('%(name)s: %(levelname)s %(message)s')
        syslog_hnd.setFormatter(formatter)
        self._logger.addHandler(syslog_hnd)
